import queue
import multiprocessing


class ProcessExecResult(object):
    """
    Результат работы процесса.
    result - результат выполнения фуникции target, заполняется только при успешном выполнении
    error_msg - экзепшен, заполняется только при генерации экзепшена из функции target
    """
    def __init__(self, result, error_msg):
        self.result = result
        self.error_msg = error_msg


class CallableProcess(multiprocessing.context.ForkProcess):
    """
    Расширенный процесс из контекста fork (multiprocessing.get_context('fork'))
    Добавляет в расширяемый процесс очередь для результатов выполненения функции target
    """

    def __init__(self, group=None, target=None, name=None, args=(), kwargs={}, result_queue={},
                 *, daemon=None):
        super().__init__(group=group, target=target, name=name, args=args, kwargs=kwargs, daemon=daemon)
        self.result_queue = result_queue

    def run(self):
        if self._target:
            try:
                result = self._target(*self._args, **self._kwargs)
                if self.result_queue:
                    self.result_queue.put(ProcessExecResult(result, None))
            except Exception as e:
                self.result_queue.popitem(ProcessExecResult(None, e))


class ProcessQueue:
    """
    Очередь задач. Служит для накопления и последующего выполнения накопленных задач.
    """
    def __init__(self, max_processes):
        self.max_processes = max_processes
        self.tasks_queue = queue.Queue()
        self.processes = []

    def sync_start(self, aggregator_func=None):
        """
        Выполнить накопленные задачи. Метод синхронный, завершится только после обработки последней задачи.
        Выполнение задач происходит паралелльно в несколько форкнутых процессов.
        При успешном выполнении процесса будет вызвана функеция aggregator_func с результатом функции taget.
        Вызов aggregator_func происходит последовательно в родительском процессе.
        """
        result_queue = multiprocessing.Queue()
        # увы, multiprocessing.Queue так работет, что результаты выполнения задач, записанные дочерними процессами
        # в эту очередь будут доступны на чтение родительскому процессу только через некоторый промежуток времени,
        # потому придется вести отдельный счетчик для гарантии что результат каждого процесса будет обработан
        processes_counter = 0
        while not self.tasks_queue.empty():
            if len(self.processes) < self.max_processes:
                task, args = self.tasks_queue.get()
                process = CallableProcess(target=task, kwargs=args, result_queue=result_queue)
                process.start()
                processes_counter += 1
            self.processes = [process for process in self.processes if process.is_alive()]
            if result_queue.qsize() != 0:
                result = result_queue.get()
                if result.result:
                    aggregator_func(result.result)
                processes_counter -= 1
        for process in self.processes:
            process.join()
            process.close()
        while processes_counter != 0:
            result = result_queue.get()
            if result.result:
                aggregator_func(result.result)
            processes_counter -= 1

    def add_task(self, task, args):
        """
        Добавить в очередь задачу.
        """
        self.tasks_queue.put((task, args))
